/*=============================================================================


    File:       opeq.pl

    Authors:    Frank Richter, Ulm University, Germany
                (based on and inspired by the confluence checker
				written by Johannes Langbein)

    Date:       26-06-2014

    Version:    1.0

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

=============================================================================*/

/*============================== chrparserOPEQ ====================================
This file is based on chrparser.pl . The following description is a modified version
of the description written by Johannes Langbein for the chrparser.pl .

The predicates in this file implement a very simple CHR parser. The parser only
works correctly on syntactically correct files and provides no means for error
handling. CHR programs should be checked for syntactical correctness by one of
the established CHR systems before they are used with this parser.

Additionally to parsing a transformed Version of the original file is created. In this
transformed version the built-ins </2,>/2,=</2,>=/2,==/2 have been replaced by CHR
constraints and constraint solvers for those built-ins have been added.

The parser ignores everything in the file expect CHR rules and the chr_constraint
directive defining the CHR constraints and their arity. This means the file can contain
arbitrary Prolog predicates, directives and CHR propagation rules as long as they are
syntactical correct.

The syntax of rules supported by this parser is given by the following EBNF
grammar.

Rule        --> [rulename '@'] ActualRule
ActualRule  --> LHS '<=>' RHS  |  LHS '==>' RHS
LHS         --> [KeptHead '\'] RemHead
RHS         --> [Guard '|'] Body
KeptHead    --> Constraints
RemHead     --> Constraints
Guard       --> Constraints
Body        --> Constraints
Constraints --> constraint {',' Constraints}

Shortly, a rule recognized by this parser is of the form
    name @ KH \ RH <=> G | B
where name, KH, and G can be omitted, together with the according symbols.
Or a rule recognized by this parser is of the form
    name @ KH ==> G | B
where name, and G can be omitted, together with the according symbols.


In the transformed file, a rule i:
rule i @ KH \ RH <=> C_built-ins,b_1(X_1_1,...,X_1_k),...,b_n(X_n_1,...,X_n_l) | B .
where C_built_ins are kept built-ins and b_1/k ... b_n/l are the replaced built-ins,
is replaced by the two rules :
rule i ask @ KH, RH ==> C_built-ins |
	ask_b_1(X_1_1,...,X_1_k , i , t(KH,RH)) ,..., ask_b_n(X_n_1,...,X_n_l , i , t(KH,RH)).
rule i entailed @ KH \ RH , entailed_b_1(X_1_1,...,X_1_k , i , T) ,..., entailed_b_n(X_n_1,...,X_n_l , i , T)
	<=> C_built-ins , T==t(KH,RH) | B .


Predicates in this file:
========================

    read_rules/4    Read rules and list of CHR constraints from a file and
					creates the transformed file.


Datastructures used:
====================

    rule(S, KH, RH, G, B)
        A term representing a CHR rule. S is the rule together with its name as
        string as it is represented in the source file. KH, RH, G, B are lists
        representing the kept head constraints, the removed head constraints,
        the guard and the body of the rule. KH, and G are possibly empty.
		For rules where built-ins in the guard have been replaced in the transformed
		file, CHR constraints representing the information needed for the guard to
		hold has been added to KH. (e.g. if A<B was present in the guard, the returend
		rule has less(A,B) in KH instead of A<B in G)

    con(F, N)
         A term representing a declaration of a CHR constraint. F is the name
         of the constraint and N is its arity.

==============================================================================*/


:- module(parserOPEQ, [read_rules/4]).

:- user:use_module(library(chr)).

%% read_rules(+FileName, +NewFile, -Rules, -CHRC)
%
% This predicate gets the path to the file it reads from as File and writes
% the transformed file to NewFile. It returns the rules of the original program
% with modified kept heads and guards, so that it is possible to extract the minimal
% states and use them with the NewFile.

read_rules(File, NewFile, Rules, CHRC) :-
    read_file(File, Clauses),
	open(NewFile,write,Stream),
	write_header(Stream),
    extract_constraint_list(Stream, Clauses, CHRC),
	write_builtin_fails(Stream,CHRC),
    extract_rules(Clauses, Stream, Rules,0),
	close(Stream).


% copy pasted from the confluence checker
% read_file(+FileName -Clauses)
%
% This predicate reads a Prolog source file provided its file name and returns
% a list of all clauses appearing in the source file. In case the source file
% contains an operator precedence declaration, this precedence is set in the
% runtime system

read_file(File,Clauses) :-
        see(File),
        read_file_aux(Clauses),
        seen.


read_file_aux([Clause|Clauses]) :-
    read(Clause),
    (Clause == end_of_file ->
         Clauses = []
    ;
         (Clause = (:- op(A,B,C)) ->
            op(A,B,C)
         ;
            true
         ),
         read_file_aux(Clauses)
    ).

% write_header(+Stream)
%
% This predicate writes a header with the rules for the constraint solvers and
% the builtin_fail constraint to the transformed file.

write_header(Stream):-
	write(Stream,':- use_module(library(chr)).'),
	nl(Stream),
	write(Stream,':- chr_constraint leq/2 , ask_leq/4 , entailed_leq/4 , less/2 , ask_less/4 , entailed_less/4 , eq/2 , ask_eq/4 , entailed_eq/4  , builtin_fail/0 .'),
	nl(Stream),
	write(Stream,'builtin_fail \\ leq(_,_) <=> true.'),
	nl(Stream),
	write(Stream,'builtin_fail \\ less(_,_) <=> true.'),
	nl(Stream),
	write(Stream,'builtin_fail \\ eq(_,_) <=> true.'),
	nl(Stream),
	write(Stream,'builtin_fail \\ ask_leq(_,_,_,_) <=> true.'),
	nl(Stream),
	write(Stream,'builtin_fail \\ ask_less(_,_,_,_) <=> true.'),
	nl(Stream),
	write(Stream,'builtin_fail \\ ask_eq(_,_,_,_) <=> true.'),
	nl(Stream),
	write(Stream,'builtin_fail \\ entailed_leq(_,_,_,_) <=> true.'),
	nl(Stream),
	write(Stream,'builtin_fail \\ entailed_less(_,_,_,_) <=> true.'),
	nl(Stream),
	write(Stream,'builtin_fail \\ entailed_eq(_,_,_,_) <=> true.'),
	nl(Stream),
	write(Stream,'redundant  @ leq(X,Y) \\ leq(X,Y)  <=> true.'),
	nl(Stream),
	write(Stream,'antisymmetric @ leq(X,Y), leq(Y,X)   <=> eq(X,Y).'),
	nl(Stream),
	write(Stream,'transitive @ leq(X,Y), leq(Y,Z)   ==> leq(X,Z).'),
	nl(Stream),
	write(Stream,'reflexive  @ leq(X,X)             <=> true.'),
	nl(Stream),
	write(Stream,'redundant  @ less(X,Y) \\ less(X,Y)  <=> true.'),
	nl(Stream),
	write(Stream,'antisymmetric @ less(X,Y), less(Y,X)   <=> builtin_fail.'),
	nl(Stream),
	write(Stream,'transitive @ less(X,Y), less(Y,Z)   ==> less(X,Z).'),
	nl(Stream),
	write(Stream,'reflexive  @ less(X,X)             <=> builtin_fail.'),
	nl(Stream),
	write(Stream,'redundant  @ eq(X,Y) \\ eq(X,Y) <=> true.'),
	nl(Stream),
	write(Stream,'transitive @ eq(X,Y), eq(Y,Z)  ==> eq(X,Z).'),
	nl(Stream),
	write(Stream,'reflexive  @ eq(X,X)             <=> true.'),
	nl(Stream),
	write(Stream,'symmetric @ eq(X,Y)    ==> eq(Y,X).'),
	nl(Stream),
	write(Stream,'eq(X,Y) <=> nonvar(X),nonvar(Y), X \\== Y | builtin_fail.'),
	nl(Stream),
	write(Stream,'eq(X,Y) , less(X,Z) ==> less(Y,Z).'),
	nl(Stream),
	write(Stream,'eq(X,Y) , less(Z,X) ==> less(Z,Y).'),
	nl(Stream),
	write(Stream,'eq(X,Y) , leq(X,Z) ==> leq(Y,Z).'),
	nl(Stream),
	write(Stream,'eq(X,Y) , leq(Z,X) ==> leq(Z,Y).'),
	nl(Stream),
	write(Stream,'less(X,Y) \\ leq(X,Y) <=> true.'),
	nl(Stream),
	write(Stream,'eq(X,Y) \\ leq(X,Y) <=> true.'),
	nl(Stream),
	write(Stream,'eq(X,Y) , less(X,Y) <=> builtin_fail.'),
	nl(Stream),
	write(Stream,'eq(X,Y) , less(Y,X) <=> builtin_fail.'),
	nl(Stream),
	write(Stream,'leq(Y,X) , less(X,Y) <=> builtin_fail.'),
	nl(Stream),
	write(Stream,'leq(X,Y)\\ask_leq(X,Y,I,T)<=>entailed_leq(X,Y,I,T).'),
	nl(Stream),
	write(Stream,'less(X,Y)\\ask_leq(X,Y,I,T)<=>entailed_leq(X,Y,I,T).'),
	nl(Stream),
	write(Stream,'eq(X,Y)\\ask_leq(X,Y,I,T)<=>entailed_leq(X,Y,I,T).'),
	nl(Stream),
	write(Stream,'eq(Y,X)\\ask_leq(X,Y,I,T)<=>entailed_leq(X,Y,I,T).'),
	nl(Stream),
	write(Stream,'ask_leq(X,X,I,T)<=>entailed_leq(X,X,I,T).'),
	nl(Stream),
	write(Stream,'ask_leq(X,Y,I,T)<=> number(X),number(Y),X=<Y| entailed_leq(X,Y,I,T).'),
	nl(Stream),
	write(Stream,'leq(X,Z)\\ask_leq(X,Y,I,T)<=> number(Z),number(Y),Z=<Y|entailed_leq(X,Y,I,T).'),
	nl(Stream),
	write(Stream,'leq(Y,X)\\ask_leq(Z,X,I,T)<=> number(Z),number(Y),Z=<Y|entailed_leq(Z,X,I,T).'),
	nl(Stream),
	write(Stream,'less(X,Z)\\ask_leq(X,Y,I,T)<=> number(Z),number(Y),Z=<Y|entailed_leq(X,Y,I,T).'),
	nl(Stream),
	write(Stream,'less(Y,X)\\ask_leq(Z,X,I,T)<=> number(Z),number(Y),Z=<Y|entailed_leq(Z,X,I,T).'),
	nl(Stream),
	write(Stream,'eq(X,Z)\\ask_leq(X,Y,I,T)<=> number(Z),number(Y),Z=<Y|entailed_leq(X,Y,I,T).'),
	nl(Stream),
	write(Stream,'eq(Y,X)\\ask_leq(Z,X,I,T)<=> number(Z),number(Y),Z=<Y|entailed_leq(Z,X,I,T).'),
	nl(Stream),
	write(Stream,'less(X,Y)\\ask_less(X,Y,I,T)<=>entailed_less(X,Y,I,T).'),
	nl(Stream),
	write(Stream,'ask_less(X,Y,I,T)<=> number(X),number(Y),X<Y| entailed_less(X,Y,I,T).'),
	nl(Stream),
	write(Stream,'less(X,Z)\\ask_less(X,Y,I,T)<=> number(Z),number(Y),Z<Y|entailed_less(X,Y,I,T).'),
	nl(Stream),
	write(Stream,'less(Y,X)\\ask_less(Z,X,I,T)<=> number(Z),number(Y),Z<Y|entailed_less(Z,X,I,T).'),
	nl(Stream),
	write(Stream,'eq(X,Y)\\ask_eq(X,Y,I,T)<=>entailed_eq(X,Y,I,T).'),
	nl(Stream),
	write(Stream,'eq(Y,X)\\ask_eq(X,Y,I,T)<=>entailed_eq(X,Y,I,T).'),
	nl(Stream),
	write(Stream,'ask_eq(X,X,I,T)<=>entailed_eq(X,X,I,T).'),
	nl(Stream),
	write(Stream,'ask_eq(X,Y,I,T)<=> X==Y| entailed_eq(X,Y,I,T).'),
	nl(Stream),
	write(Stream,'eq(X,Z)\\ask_eq(X,Y,I,T)<=> Z==Y|entailed_eq(X,Y,I,T).'),
	nl(Stream),
	write(Stream,'eq(Y,X)\\ask_eq(Z,X,I,T)<=> Z==Y|entailed_eq(Z,X,I,T).'),
	nl(Stream).

% copy pasted from the confluence checker
% extract_rules(+Clauses, +Stream, -Rules, +I)
%
% Given a list of Prolog clauses, this predicate returns a list of rules in
% the format described above. This predicate ignores all clauses
% which are not built according to the aforementioned grammar.
%
% modification: I has been added to know the number of the curent rule and it gets
% incremented for the next call of the predicate. Stream has been added to be able
% to write to the Stream which just passed through for other predicates that need them.

extract_rules([end_of_file|_], _, [],_) :- !.

extract_rules([H|R], Stream, [Rule|Rules],I) :-
    parse_rule(I,H, Stream, Rule),
    !,
	Inew is I + 1,
    extract_rules(R, Stream, Rules,Inew).

extract_rules([_|R], Stream, Rules,I) :-
    extract_rules(R, Stream, Rules,I).

% copy pasted from the confluence checker
% parse_rule(I, +Term, Stream, -Rule)
%
% This predicate succeeds if Term is a CHR rule according to the aforementioned
% grammar. In this case, Rule is unified with a term of the form
% rule(N, KH, RH, G, B) representing the rule.
%
% modification: I has been added to know the number of the curent rule, and Stream
% has been added to be able to write to the Stream. Both are just passed through
% for other predicates that need them.


parse_rule(I,Term, Stream, rule(Term, KH, RH, G, B)) :-
    Term = (_@ActRule),
    !,
    parse_actrule(I,Stream, ActRule, KH, RH, G, B).

parse_rule(I,Term, Stream, rule(Term, KH, RH, G, B)) :-
    parse_actrule(I,Stream, Term, KH, RH, G, B).



% parse_actrule(+I,+Stream,+Term, -KH, -RH, -G, -B)
%
% (description from chrparser.pl:)
% This predicate succeeds if Term is constructed according to the grammar rule
% for ActualRule in the aforementioned grammar. In this case, KH and RH are
% unified with the list of kept and removed head constraints respectively,
% while G and B are unified with the lists of constraints from the guard and
% body of the rule, respectively.
%
% This predicate additionally calls the predicates to transform the rule and
% write it to the transformed file.
% I has been added to know the number of the current rule, and Stream
% has been added to be able to write to the Stream. Both are just passed through
% for other predicates that need them.

parse_actrule(I,Stream, Term, KHnew, RH, Gnew, B) :-
    Term = (LHS==>RHS),
    !,
    parse_lhs_prop(LHS, KH, RH),
    parse_rhs(RHS, G, B),
	check_guard_leq(G,Gnewleq,Gleq),
	check_guard_less(Gnewleq,Gnewless,Gless),
	check_guard_eq(Gnewless,Gnew,Geq),
	append(Gleq,Gless,Gleqless),
	append(Gleqless,Geq,Ghead),
	create_entailed(I,T,Ghead,Ent),
	append(KH,Ghead,KHnew),
	append(RH,Ent,RHnew),
	list2goal(KH,KHf),
	list2goal(RHnew,RHf),
	list2goal(B,Bf),
	append(KH,RH,Head),
	list2goal(Head,Headf),
	create_new_guard(Ghead,Headf,Gnew,Gnewent,T),
	list2goal(Gnewent,Gf),
	write_ask(Stream, Headf, Ghead,I),
	write_rule(KHf,RHf,Gf,Bf,Stream).

parse_actrule(I,Stream, Term, KHnew, RH, Gnew, B) :-
    Term = (LHS<=>RHS),
    !,
    parse_lhs(LHS, KH, RH),
    parse_rhs(RHS, G, B),
	check_guard_leq(G,Gnewleq,Gleq),
	check_guard_less(Gnewleq,Gnewless,Gless),
	check_guard_eq(Gnewless,Gnew,Geq),
	append(Gleq,Gless,Gleqless),
	append(Gleqless,Geq,Ghead),
	create_entailed(I,T,Ghead,Ent),
	append(KH,Ghead,KHnew),
	append(RH,Ent,RHnew),
	list2goal(KH,KHf),
	list2goal(RHnew,RHf),
	list2goal(B,Bf),
	append(KH,RH,Head),
	list2goal(Head,Headf),
	create_new_guard(Ghead,Headf,Gnew,Gnewent,T),
	list2goal(Gnewent,Gf),
	write_ask(Stream, Headf, Ghead,I),
	write_rule(KHf,RHf,Gf,Bf,Stream).



create_new_guard([],_,G,G,_).

create_new_guard(_,Head,G,Gent,T) :- append(G,[(T==t(Head))],Gent).

create_entailed(_,_,[],[]).

create_entailed(I,T,[leq(A,B)|Gs],[entailed_leq(A,B,I,T)|Es]):-
	create_entailed(I,T,Gs,Es).

create_entailed(I,T,[less(A,B)|Gs],[entailed_less(A,B,I,T)|Es]):-
	create_entailed(I,T,Gs,Es).

create_entailed(I,T,[eq(A,B)|Gs],[entailed_eq(A,B,I,T)|Es]):-
	create_entailed(I,T,Gs,Es).


% list2goal
%
% This predicate is similar to the list2goal from conflcheckerhelp.pl that converts
% a list of predicates to a list of goals that can be used with call. The difference is,
% that for an empty list it returns the goal 'true'.

list2goal([],G) :-
	G = true.

list2goal([X],X).

list2goal([X|Xs],G) :-
	G=(X,GXs), list2goal(Xs,GXs).


% write_rule(+KH,+RH,+G,+B,Stream)
%
% This predicate writes a rule to the transformed file

write_rule(true,RH,Gnew,B,Stream) :-
	write(Stream,RH<=>Gnew|B),write(Stream,'.'),nl(Stream).

write_rule(KH,true,Gnew,B,Stream) :-
	write(Stream,KH==>Gnew|B),write(Stream,'.'),nl(Stream).

write_rule(KH,RH,Gnew,B,Stream) :-
	write(Stream,KH\RH<=>Gnew|B),write(Stream,'.'),nl(Stream).


% write_ask(+Stream,+Head,+G,+I)
%
% This predicate writes a rule that adds a ask rule to the
% transformed file if the original rule has built-ins in the guard
% that are removed during the transformation.

write_ask(_,_,[],_).

write_ask(Stream,Head,Gs,I) :-
	write(Stream,Head),
	write(Stream,'==>'),
	write_ask1(Stream,Head,Gs,I),
	write(Stream,'.'),
	nl(Stream).

% write_ask1(+Stream,+Head,+G,+I)
%
% This predicate writes the first ask constraint for a rule and
% calls a predicate to write the rest of the ask constraints

write_ask1(Stream,Head,[leq(A,B)|Gs],I) :-
	write(Stream,ask_leq(A,B,I,t(Head))),
	write_ask_rest(Stream,Head,Gs,I).

write_ask1(Stream,Head,[less(A,B)|Gs],I) :-
	write(Stream,ask_less(A,B,I,t(Head))),
	write_ask_rest(Stream,Head,Gs,I).

write_ask1(Stream,Head,[eq(A,B)|Gs],I) :-
	write(Stream,ask_eq(A,B,I,t(Head))),
	write_ask_rest(Stream,Head,Gs,I).


% write_ask_rest(+Stream,+Head,+G,+I)
%
% This predicate writes the rest of the ask constraints separated
% by ',' to match the syntax.

write_ask_rest(_,_,[],_).

write_ask_rest(Stream,Head,[leq(A,B)|Gs],I) :-
	write(Stream,','),
	write(Stream,ask_leq(A,B,I,t(Head))),
	write_ask_rest(Stream,Head,Gs,I).

write_ask_rest(Stream,Head,[less(A,B)|Gs],I) :-
	write(Stream,','),
	write(Stream,ask_less(A,B,I,t(Head))),
	write_ask_rest(Stream,Head,Gs,I).

write_ask_rest(Stream,Head,[eq(A,B)|Gs],I) :-
	write(Stream,','),
	write(Stream,ask_eq(A,B,I,t(Head))),
	write_ask_rest(Stream,Head,Gs,I).

% copy pasted from the confluence checker
% parse_lhs(+LHS, -KH, -RH)
%
% This predicate succeeds, if LHS is the left-hand side of a CHR rule according
% to the above grammar. In this case, KH is the list of kept head constraints,
% RH is the list of removed head constraints. If the left-hand side of a
% simplification rule is parsed, KH is the empty list.

parse_lhs(LHS, KH, RH) :-
    LHS = (KHS\RHS),!,
    parse_constraints(KHS, KH),
    parse_constraints(RHS, RH).

parse_lhs(LHS, [], RH) :-
    parse_constraints(LHS, RH).

%added for propagation rules
parse_lhs_prop(LHS, KH, []) :-
    parse_constraints(LHS, KH).


% copy pasted from the confluence checker
% parse_rhs(+RHS, -G, -B)
%
% This predicate succeeds, if RHS is the right-hand side of a CHR rule
% according to the above grammar. In this case, G is the list of the built-in
% constraints from the guard of the rule and B is the list containing the
% constraints from the body of the rule. G is the empty list if the rule
% contains no guard.


parse_rhs(RHS, G, B) :-
    RHS = (GS|BS),
    !,
    parse_constraints(GS, G),
    parse_constraints(BS, B).

parse_rhs(RHS, [], G) :-
    parse_constraints(RHS, G).


% copy pasted from the confluence checker
% parse_constraints(+Term, -List)
%
% This predicate succeeds if term is an arbitrary term or a sequence of terms,
% separated by (,). In this case List is the list containing all the terms
% from the sequence. This predicate is used to parse the head, guard, or body
% of a rule into a list of constraints.

parse_constraints(Term, [C|Res]) :-
    Term = (C,CS),
    !,
    parse_constraints(CS, Res).

parse_constraints(Term, [Term]).


% check_guard_leq(+G,-Gnew,-Ghead)
%
% This predicate checks a guard for =< , removes them form the guard
% and adds them in the form of leq constraints to a list that
% needs to be added to the kept head

check_guard_leq([],[],[]).

check_guard_leq([G|Gs],Gnew,[Gh|Ghs]):-
	G =(A=<B),
	!,
	Gh = leq(A,B),
	check_guard_leq(Gs,Gnew,Ghs).

check_guard_leq([G|Gs],Gnew,[Gh|Ghs]):-
	G =(B>=A),
	!,
	Gh = leq(A,B),
	check_guard_leq(Gs,Gnew,Ghs).

check_guard_leq([G|Gs],[G|Gns],Ghead):-
	!,
	check_guard_leq(Gs,Gns,Ghead).


% check_guard_less(+G,-Gnew,-Ghead)
%
% This predicate is like check_guard_leq with < instead of =< and
% the less constraint instead of the leq constraint.

check_guard_less([],[],[]).

check_guard_less([G|Gs],Gnew,[Gh|Ghs]):-
	G =(A<B),
	!,
	Gh = less(A,B),
	check_guard_less(Gs,Gnew,Ghs).

check_guard_less([G|Gs],Gnew,[Gh|Ghs]):-
	G =(B>A),
	!,
	Gh = less(A,B),
	check_guard_less(Gs,Gnew,Ghs).

check_guard_less([G|Gs],[G|Gns],Ghead):-
	!,
	check_guard_less(Gs,Gns,Ghead).


% check_guard_eq(+G,-Gnew,-Ghead)
%
% This predicate is like check_guard_leq with == instead of =< and
% the less constraint instead of the eq constraint.

check_guard_eq([],[],[]).

check_guard_eq([G|Gs],Gnew,[Gh|Ghs]):-
	G =(A==B),
	!,
	Gh = eq(A,B),
	check_guard_eq(Gs,Gnew,Ghs).

check_guard_eq([G|Gs],[G|Gns],Ghead):-
	!,
	check_guard_eq(Gs,Gns,Ghead).


% write_builtin_fails(+C)
%
% This predicate writes rules to remove user defined CHR constraints C
% in the case that a builtin_fail constraint is present.

write_builtin_fails(_,[]).

write_builtin_fails(Stream,[(C/0) | T]) :-
	write(Stream,'builtin_fail \\ '),
	write(Stream,C),
	write(Stream,' <=> true.'),
	nl(Stream),
	write_builtin_fails(Stream,T).

write_builtin_fails(Stream,[(C/X) | T]) :-

	write(Stream,'builtin_fail \\ '),
	write(Stream,C),
	write(Stream,'(_'),
	Z is X - 1 ,
	create_wildcards(Stream,Z),
	write(Stream,') <=> true.'),
	nl(Stream),
	write_builtin_fails(Stream,T).


% create_wildcards(Stream,X)
%
% This predicate creates wildcards for the rules that remove user
% defined CHR constraints. It writes X times ',_' to the Stream.

create_wildcards(_,0).

create_wildcards(Stream,X) :-
	write(Stream,',_'), Y is X -1 , create_wildcards(Stream,Y).

% copy pasted from the confluence checker
% extract_constraint_list(+Clauses ~CHRCons)
%
% This predicate takes a list of read-in Prolog clauses and extracts the list
% of CHR constraints as defined by the directive ":- chr_constraints..." in
% the source file. The resulting list CHRCons consists of terms of the form
% (F, N), where F is the functor and N the arity of a CHR constraint as
% defined by the directive. If no such directive is found in the list Clauses,
% the predicate fails.

extract_constraint_list(_, [], []).

extract_constraint_list(Stream, [C|R], CHRCnew) :-
    C = (:- chr_constraint CS),
    !,
	write(Stream,':- chr_constraint '),
	write(Stream, CS),
	write(Stream, '.'),
	nl(Stream),
    parse_constraint_list(CS, CHRC),
	append(CHRC,CHRCrest,CHRCnew),
	extract_constraint_list(Stream, R, CHRCrest).

extract_constraint_list(Stream, [_|R], CHRC) :-
    extract_constraint_list(Stream, R, CHRC).

% copy pasted from the confluence checker
% parse_constraint_list(+ConList, ~CHRCons)
%
% This predicate parses the term ConList of the form (F1/N1, F2/N2, ...) where
% the Fi's are functors and the Ni's are arities, into a list CHRCons of terms
% of the form (Fi, Ni).

parse_constraint_list(CL, [(F/A)|CHRC]) :-
    CL = (C,R),
    C = (F/A),!,
    parse_constraint_list(R, CHRC).

parse_constraint_list(C, [(F/A)]) :- C = (F/A).