/*=============================================================================


    File:       opeq.pl

    Authors:    Frank Richter, Ulm University, Germany

    Date:       26-06-2014
    
    Version:    1.0 

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

=============================================================================*/

/*============================== cleanup ====================================

This file is used to clean up a final state from any leftover ask or entailed
constraints. It also unifies any free variables in eq constraints to ensure that
the state equivalence check does consider those Variables equivalent.


Predicates in this file:
========================

    cleanup/2    removes leftover ask and entailed constraints
    
    
==============================================================================*/

:- module(cleanup, [cleanup/2]).
	
% cleanup(+Constraints,-Cleaned)
%
% This predicate gets a list of constraints as Constraints and returns a list
% of constraints as Cleaned. The returned list is a filtered version of the
% Constraints list where all ask_leq/4, entailed_leq/4, ask_less/4 ,entailed_less/4
% ask_eq/4 and entailed_eq/4 constraints have been removed. If this predicates
% encounters any eq constraints it calls the predicate unifyeq/2 to ensure that
% the free variables are unified.
	
cleanup([],[]).
cleanup([(ask_leq(_,_,_,_))|Xs],Ys):- cleanup(Xs,Ys).
cleanup([(entailed_leq(_,_,_,_))|Xs],Ys):- cleanup(Xs,Ys).
cleanup([(ask_less(_,_,_,_))|Xs],Ys):- cleanup(Xs,Ys).
cleanup([(entailed_less(_,_,_,_))|Xs],Ys):- cleanup(Xs,Ys).
cleanup([(ask_eq(_,_,_,_))|Xs],Ys):- cleanup(Xs,Ys).
cleanup([(entailed_eq(_,_,_,_))|Xs],Ys):- cleanup(Xs,Ys).
cleanup([(eq(A,B))|Xs],[(eq(A,B))|Ys]):- unifyeq(A,B),unifyeq(B,A),cleanup(Xs,Ys).
cleanup([X|Xs],[X|Ys]):- cleanup(Xs,Ys).

% unifyeq(+A,+B)
%
% This predicate checks if A is a free variable and unifies it with B if it is.
% If A is not a free Variable this predicate succeeds without doing anything.

unifyeq(A,B):-var(A),A=B.
unifyeq(_,_).