/*=============================== conflcheckehelp ===========================

This file is the rest of the conflcheck.pl file, the original file was split into
two parts because some predicates of it were needed in a module that is loaded by
the conflcheck.pl .

IMPORTANT NOTE:
    The predicates in this file write temporary files to your hard-drive. Thus, 
    it needs to be executed in a directory with write access. All files are 
    deleted after the execution.



Predicates in this file:
========================

	list_to_goal/2				converts a list to a goal
	
	create_file_name/1			creates a random file name that is not used by another file
	
	findall_chr_constraints/2	returns all CHR constraints that are present in the constraintstore
	
	reconnect/6					reconnects the global variables


Datastructures used:
====================

    state(G, B, V)
        A term representing a CHR state. 
        G is a list of CHR constraints, represented as Prolog terms. 
        B is a list of built-in constraints, represented as Prolog terms.
        V is the list of global variables, represented as unbound Prolog 
        variables. This list must not contain duplicates.
        
    rule(S, KH, RH, G, B)
        A term representing a CHR rule. S is the rule together with its name as 
        string as it is represented in the source file. KH, RH, G, B are lists 
        representing the kept head constraints, the removed head constraints, 
        the guard and the body of the rule. KH, and G are possibly empty. 

    cp(S1, S2, R1, R2, O, CAS)
        A term representing a critical pair. 
        S1 and S2 are the two states of the critical pair represented as 
        state(...) terms according to the definition above. 
        R1 and R2 are the CHR rules the critical pair is stemming from. They 
        are represented as rule(...) terms according to the definition above. 
        O is a list of tuples, consisting of pairs of constraints. Those pairs 
        describe the overlap of the two rules which leads to the critical pair. 
        The list [(a(X),a(Y)), (b(X),b(Y))] means, that the head constraint 
        a(X) of the first rule is overlapped with the head constraint a(Y) of 
        the second rule and also the head constraint b(X) of the first rules is 
        overlapped with the head constraint b(Y) of the second rule. 
        The list CAS contains the critical ancestor state of this critical pair.

=============================================================================*/


:- module(conflcheckhelp, 	[ list_to_goal/2,
							findall_chr_constraints/2,
							create_file_name/1,
							reconnect/6]).
							
:- use_module(termlib, [tl_member/3, tl_filter/3, tl_make_set/3]).

% list_to_goal(+L, ~G)
%
% This predicate converts a list of constraints L (user-defined as well as
% built-ins constraints) into a single goal which then can be executed using
% the predicate call/1.

list_to_goal([C], (C)).
list_to_goal([C|R],(C,T)) :- list_to_goal(R, T).


% findall_chr_constraints(+GlobVars, -Res)
%
% This predicate binds Res to the list of all CHR constraints in the store that 
% can be found using find_chr_constraint/1 from the CHR library. The list 
% GlobVars contains all the global variables appearing in the original query 
% which led to the result Res. This is needed to keep the global variables 
% connected to the result of the computation.
%
% This predicate temporarly save the CHR constraints into a file to keep the
% variables connected. Thus it needs to be placed in a directory where the user 
% has write access. The filename is randomly generated and the file is deleted 
% after the computation.

findall_chr_constraints(GlobVars, Res) :-
    create_file_name(Name),
    open(Name, write, Sout, [alias(output)]),
    (
        (
            find_chr_constraint(R),
            write(output, R),
            write(output, ', '),
            nl(output),
            fail
        )
        ;
        (
            true
        )
    ),
    write(output, globs(GlobVars)),
    write(output, '.'),
    nl(output),
    close(Sout),
    open(Name, read, Sin, [alias(input)]),
    read(input, T), 
    close(Sin),
    delete_file(Name),
    to_list(T, Res).


% to_list(+Term, -List)
%
% This predicate converts the conjunction of terms T, ending with a globs(...) 
% term into a list, where each subterm is one element fo the list.
%
% Example: to_list((a(X),b(y,(c(Z)),globs([X,Y,Z])), R), yields 
% R = [a(X), b(y,(c(Z)), globs([X,Y,Z])]

to_list(T, [H|Res]) :- 
    T =.. [_,H,T1],
    to_list(T1, Res).

to_list(T, [T]) :- 
    T =.. [globs, _].


% create_file_name(-Name)
% 
% This predicate randomly generates a filename. If a file with this name does 
% not exist in the execution directory, the filename is returned, otherwise a 
% different name is generated
create_file_name(Name) :-
    random(X),
    string_to_atom(Name, X),
    (exists_file(Name) -> fail; true).


% reconnect(+Res1, +Res2, -Goal1, -Goal2, -GlobVars1, -GlobVars2)
%
% This predicate seperates the results found by findall_chr_constraints/2 into
% the actual goal store and the set of global variables. Furthermore, it
% reconnects the two states by unifiying the global variables of the two states,
% while keeping track, that a global variable from one state is only bound to 
% one global variable from another state (for details see reconnect_globs/3).
% The lists Res1 and Res2 are the results found by findall_chr_constraints/2. 
% Goal1 and Goal2 are lists representing the respective goal stores while 
% GlobVars1 and GlobVars2 are lists containing the global variables.

reconnect(R1, R2, G1, G2, V1, V2):-
    extract_globs(R1, G1, V1),
    extract_globs(R2, G2, V2),
    reconnect_globs(V1, V2).


% extract_globs(+List, -Goal, -GlobVars)
%
% This predicate separates the list List generated by findall_chr_constraints/2
% into the goal store Goal and the list of global variables GlobVars.

extract_globs([globs(V)], [], V).

extract_globs([H|R], [H|G], V) :-
    extract_globs(R, G, V).


% reconnect_globs(+V1, +V2)
%
% Auxilliary predicate, calls reconnect_globs(V1, V2, []).

reconnect_globs(V1, V2) :- reconnect_globs(V1, V2, []).


% reconnect_globs(+V1, +V2, S)
%
% This predicate reconnect the lists of global variables V1 and V2. This needs 
% to be done if the global variables originate from two states of a critical 
% pair that have been executed separately and thus have been disconnected. The 
% predicate makes sure, that every variable occuring in V1 is only unified with 
% one variable occuring in V2 and vice versa. This is needed to take into 
% account different variable bindings that may have occured during execution 
% of the states of the critical pair. If an already reconnected variable is 
% seen again, it needs to be bound to the same variable as before, otherwise 
% the predicate fails. S is a buffer variable wich needs to be instantiated to 
% the empty list when calling the predicate.
%
% Example 1:
% reconnect_globs([X, Y, X], [A, B, A],[]) succeeds with X = A and Y = B.
%
% Example 2:
% reconnect_globs([X, Y, X], [A, B, C],[]) fails.

reconnect_globs([], [], _).

reconnect_globs([V1|R1],[V2|R2], S) :-
    tl_member(V1, S, ==),!,
    V1 == V2,
    reconnect_globs(R1, R2, S).

reconnect_globs([V1|R1],[V2|R2], S) :-
    tl_member(V2, S, ==),!,
    V1 == V2,
    reconnect_globs(R1, R2, S).

reconnect_globs([V1|R1],[V2|R2], S) :-
    \+ tl_member(V1, S, ==),
    \+ tl_member(V2, S, ==),
    var(V1),
    var(V2),
    V1 = V2,
    reconnect_globs(R1, R2, [V1|S]).
    
reconnect_globs([V1|R1],[V2|R2], S) :-
    compound(V1),
    compound(V2),
    V1 =..[_|T1],
    V2 =..[_|T2],
    append(T1, R1, L1),
    append(T2, R2, L2),
    reconnect_globs(L1, L2, S).
    
reconnect_globs([V1|R1],[V2|R2], S) :-
    ground(V1),
    ground(V2),
    V1 == V2,
    reconnect_globs(R1, R2, S).