/*=============================================================================


    File:       opeq.pl

    Authors:    Frank Richter, Ulm University, Germany
                (based on and inspired by the confluence checker
				written by Johannes Langbein)

    Date:       26-06-2014
    
    Version:    1.0 

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

=============================================================================*/

/*=============================== confluence check ===========================

The predicates in this file implement a operational equivalence check based on Theorem 5.27
from the Constraint Handling Rules Book page 128 (ISBN: 978-0-521-87776-3). 

The predicate opeq/2 compares two well-behaved programs while the predicate confopeq/2 takes 
two terminating programs that are not allowed to have propagation rules. Both predicates only 
support the built-ins true/0, =/2, </2, >/2, =</2, >=/2, ==/2.

IMPORTANT NOTE:
    The predicates in this file write temporary files to your hard-drive. Thus, 
    it needs to be executed in a directory with write access. All files are 
    deleted after the execution.


Predicates in this file:
========================

	opeq/2			checks two programs for operational equivalence
	
	confopeq/2		checks two programs for confluence and operational equivalence


Datastructures used:
====================

	minState(G,B,V)
		A term representing the minimal state of a rule.
		It is the same as state(G, B, V) described in the following.

%copy pasted from the conflcheck.pl:

    state(G, B, V)
        A term representing a CHR state. 
        G is a list of CHR constraints, represented as Prolog terms. 
        B is a list of built-in constraints, represented as Prolog terms.
        V is the list of global variables, represented as unbound Prolog 
        variables. This list must not contain duplicates.
        
    rule(S, KH, RH, G, B)
        A term representing a CHR rule. S is the rule together with its name as 
        string as it is represented in the source file. KH, RH, G, B are lists 
        representing the kept head constraints, the removed head constraints, 
        the guard and the body of the rule. KH, and G are possibly empty. 

=============================================================================*/

:- module(opeq, 	[ opeq/2,
					confopeq/2]).

:- use_module(chrparserOPEQ, [read_rules/4]).
:- use_module(conflcheck, [check_confluence1/4, check_confluence/1]).
:- use_module(stateequiv, [equivalent_states/2]).
:- use_module(termlib, [tl_member/3]).
:- use_module(cleanup, [cleanup/2]).
:- use_module(consult1, [consult_file1/4]).
:- use_module(consult2, [consult_file2/4]).
:- use_module(chrlib, [builtins_consistent/1, unifier/3]).
:- use_module(conflcheckhelp, 	[ list_to_goal/2,
							findall_chr_constraints/2,
							create_file_name/1,
							reconnect/6]).


% opeq(+File1,+File2)
%
% This predicate checks if File1 and File2 are operational equivalent.
% File1 and File2 are strings containing the complete path of both input files.
% First two random unused filenames are generated to store the transformed code 
% of the tested files. Next read_rules/4 is called for both test files this returns 
% a list of all the rules in the file, a list of all the constraints and creates 
% the transformed code of the file. The two lists of constraints are checked by
% check_Chrc/3 and the operational equivalence check is continued in the predicate
% continueopeq/5

opeq(File1,File2):- 
	create_file_name(FileExe1),
	create_file_name(FileExe2),
	read_rules(File1,FileExe1,Prog1,Chrc1),
	read_rules(File2,FileExe2,Prog2,Chrc2),
	check_Chrc(Chrc1,Chrc2,N),!,
	continueopeq(Prog1,Prog2,FileExe1,FileExe2,N).
	

% continueopeq(+Prog1,+Prog2,+File1t,+File2t,+N)
%
% This predicate continues the operational equivalence check after the check of the
% two constraint lists. If N = 1 it means that at least one constraint of a test file
% is not present in the other test file and no further check needs to be done.
% If N = 0 both test files have the same constraints defined and the list of minimal
% states for all rules of both files with the predicate list_of_min_states/3.
% The predicate opeqtest/4 is called next to perform a check for each minimal state.
% After those checks the temporary files with the transformed code are no longer 
% needed, so the get deleted by delete_file/1. Finally the result of the operational
% equivalence check is printed on the screen by print_result/1.
	
continueopeq(_,_,_,_,1):-!.
	
continueopeq(Prog1,Prog2,File1t,File2t,0):-
	list_of_min_states(Prog1,Prog2, Min_States),!,
	opeqtest(Min_States,File1t,File2t,N),
	delete_file(File1t),
	delete_file(File2t),
	print_result(N).
	
% confopeq(+File1,+File2)
%
% The beginning of this predicate is equivalent to opeq/2, but after the rules are read
% a check_confluence1/4 is called for each test file to check if it is confluent. The
% number of not joinable critical pairs of both programs is added up and the check is
% continued with a call of confopeqcontinue/7.
	
confopeq(File1,File2):-
	create_file_name(FileExe1),
	create_file_name(FileExe2),
	read_rules(File1,FileExe1,Prog1,Chrc1),
	read_rules(File2,FileExe2,Prog2,Chrc2),
	check_confluence1(File1,FileExe1,1,N1),
	check_confluence1(File2,FileExe2,2,N2),
	N is N1 + N2,!,
	confopeqcontinue(Prog1,Chrc1,Prog2,Chrc2,FileExe1,FileExe2,N).

% confopeqcontinue(+Prog1,+Chrc1,+Prog2,+Chrc2,+FileExe1,+FileExe2,+N)
%
% If N = 0 the check is continued like in opeq/2.
% If N > 0 it means that at least one of the test files is not confluent and no further
% checks will be done. The temporary files with the transformed code are deleted and a
% message that the test failed is printed on the screen.
	
confopeqcontinue(Prog1,Chrc1,Prog2,Chrc2,File1t,File2t,0):-
	check_Chrc(Chrc1,Chrc2,N),!,
	continueopeq(Prog1,Prog2,File1t,File2t,N).
	
confopeqcontinue(_,_,_,_,File1t,File2t,N) :- 
	N>0 , 
	delete_file(File1t),
	delete_file(File2t),
	print('The input programs are not confluent.\n').
	
% check_Chrc(+Chrc1,+Chrc2,-N)
%
% This predicate gets two lists and checks if both have the exact same elements 
% (not necessarily in the same order) and returns N = 0 if they do or N = 1 if they don't.

	
check_Chrc(Chrc,Chrc,0) :- !.

check_Chrc(Chrc1,Chrc2,N) :-
	((check_Chrc_members(Chrc1,Chrc2),
	check_Chrc_members(Chrc2,Chrc1))
	->
        (N is 0)
	;
        ((N is 1),
		print('The two programs have different CHR constraints defined.'))
    ).

% check_Chrc_members(+Chrc1,+Chrc2)	
%
% This predicate gets two lists and succeeds if all elements of the first list are 
% also present in the second list.
	
check_Chrc_members([],_).
check_Chrc_members([C|Cs],Chrc2):-
	member(C,Chrc2),!,
	check_Chrc_members(Cs,Chrc2).

% print_result(+N)
%
% This predicate gets the number of minimal states that lead to different final states and
% prints the result of the operational equivalence test accordingly.	
	
print_result(0) :- 
	print('\n\nThe programs are operationally equivalent.').
	
print_result(N) :- 
	print('\n\nThe programs are not operationally equivalent, \n'),
	print(N),
	print(' minimal state(s) are not joinable.').

% list_of_min_states(+Prog1,+Prog2,-Min_States)
%
% This predicate executes list_of_min_states/2 for the list of rules of each input file and appends the lists.
	
list_of_min_states(Prog1,Prog2, Min_States) :- 
	list_of_min_states(Prog1,Min_States1),
	list_of_min_states(Prog2,Min_States2),
	append([Min_States1,Min_States2],
	Min_States).
	
% list_of_min_states(+Prog,-Min_States)
%
% This predicate extracts all minimal states from the rules.

list_of_min_states([],[]).
list_of_min_states([rule(_, KH, RH, G, _)|B],[minState(C,G,V)|R]) :- 
	append([KH,RH],C),
	append([C,G],V1),
	term_variables(V1,V),
	list_of_min_states(B,R).


% opeqtest(+Min_States,+File1,+File2)
%
% This predicate calls states_joinable/4 with each minimal state of the list of minimal states.
% It returns the number of minimal states that lead to different final states as N.

opeqtest([],_,_,0).
opeqtest([S|Ss],F1,F2,N):- !,
	states_joinable(S, F1, F2, N1),
	opeqtest(Ss,F1,F2,Nnew),
	N is Nnew + N1.

% states_joinable(+S, +File1, +File2, -N)
% 
% This predicate consults File1 and File2 with the minimal state S, calls cleanup/2 to 
% remove leftover ask or entailed constraints and checks if the final states are equivalent.
% The way this check is performed is a modified version of the states_joinable that can be 
% found in conflcheck.pl.

states_joinable(S, File1, File2 , N) :-
    copy_term(S, minState(G1, B1, V1)),
    copy_term(S, minState(G2, B2, V2)),
    append(G1, B1, L1),
    append(G2, B2, L2),
    list_to_goal(L1, H1),
    list_to_goal(L2, H2), 
	set_prolog_flag(verbose, silent),
	style_check(-singleton),
	set_prolog_flag(chr_toplevel_show_store, false),!,
	consult_file1(File1,H1,V1,Result1),
	cleanup(Result1,Result1clean),!,
	consult_file2(File2,H2,V2,Result2),
	cleanup(Result2,Result2clean),
	set_prolog_flag(chr_toplevel_show_store, true),
    style_check(-singleton),
    set_prolog_flag(verbose, normal),!,
	eq_states(Result1clean,Result2clean,N,S).
   
   
% eq_states(+Result1clean,+Result2clean,-N,+S)
%
% This predicate calls the equivalent_states/2 predicate for the two states it is called with.
% If either the equivalent_states/2 succeeds or the final state has the builtin_fail/0 constraint
% this predicate returns N = 0, else it returns N = 1 and prints the minimal state and the two
% not joinable final states to the screen.

eq_states(Result1clean,Result2clean,N,S):-
	(((reconnect(Result1clean, Result2clean, G1n, G2n, V1n, V2n), equivalent_states(state(G1n, [], V1n), state(G2n, [], V2n)));
	(member((builtin_fail),Result1clean),member((builtin_fail),Result2clean)))
	->
        (N is 0)
	;
        ((N is 1),
		numbervars(states(S,Result1clean , Result2clean), 0, _E), 
		write_term('=====================================================================', []),nl,
		print('For the minimal State:\n'),
		write_term(S, [numbervars(true)]), nl, 
		print('\nfile 1 results in:\n'),
		write_term(Result1clean, [numbervars(true)]), nl, 
		print('\nwhile file 2 results in: \n'),
		write_term(Result2clean, [numbervars(true)]), nl, 
		print('\nthese states are not joinable \n')),
		write_term('=====================================================================', []),nl
    ).